import WebDavService from "@/services/WebDavService.js";
import { WebDavStore } from "@/stores/WebDavStore.js";
import { RewriteItemsData } from "@/models/RewriteItemsData.js";
import CommonService from "@/services/CommonService";
import { WebDavError } from "@/models/WebDavError.js";
import { ITHit } from "webdav.client";
import i18n from "@/i18n";
export class UploadItemRow {
  UploadItem = null;
  CurrentState = null;
  RetryMessage = null;
  constructor(uploadItem) {
    this.UploadItem = uploadItem;
    this.UploadItem.AddListener("OnProgressChanged", this._OnProgress);
    this.UploadItem.AddListener("OnStateChanged", this._OnStateChange);
    this.UploadItem.AddListener(
      "OnBeforeUploadStarted",
      this._OnBeforeUploadStarted
    );
    this.UploadItem.AddListener("OnUploadError", this._OnUploadError);
    this._MaxRetry = 10;
    this._CurrentRetry = 0;
    this._RetryDelay = 10;
    this.CurrentState = this.UploadItem.GetState();
  }

  _OnProgress = oProgressEvent => {
    this.CurrentState = oProgressEvent.Sender.GetState();
  };

  _OnStateChange = oStateChanged => {
    this.CurrentState = oStateChanged.NewState;
    if (oStateChanged.NewState === ITHit.WebDAV.Client.Upload.State.Completed) {
      WebDavStore.commit("updateGrid");
    }
  };

  _OnBeforeUploadStarted = oBeforeUploadStarted => {
    var oItem = oBeforeUploadStarted.Sender;
    var sHref = ITHit.EncodeURI(oItem.GetUrl());
    if (
      oItem.GetOverwrite() ||
      oItem.IsFolder() ||
      oItem.CustomData.FileExistanceVerified
    ) {
      oBeforeUploadStarted.Upload();
      return;
    }
    WebDavService.WebDavSession.OpenItemAsync(sHref, [], function(
      oAsyncResult
    ) {
      if (!oAsyncResult.IsSuccess && oAsyncResult.Status.Code === 404) {
        // The file does not exist on the server, start the upload.
        oBeforeUploadStarted.Upload();
        return;
      }

      if (!oAsyncResult.IsSuccess) {
        WebDavStore.commit(
          "setError",
          new WebDavError(
            i18n.t("phrases.errors.failedCheckExistsErrorMessage"),
            oAsyncResult.Error
          )
        );
        oBeforeUploadStarted.Sender.SetFailed(oAsyncResult.Error);
        return;
      }

      var rewriteData = new RewriteItemsData(
        /* A user selected to overwrite existing file. */
        function() {
          // Do not delete item if upload canceled (it existed before the upload).
          oBeforeUploadStarted.Sender.SetDeleteOnCancel(false);

          // The item will be overwritten if it exists on the server.
          oBeforeUploadStarted.Sender.SetOverwrite(true);

          // All async requests completed - start upload.
          oBeforeUploadStarted.Upload();
        },
        /* A user selected to skip existing files. */
        function() {},
        oItem.GetRelativePath()
      );
      WebDavStore.commit("setRewriteItemsData", rewriteData);
    });
  };
  _CancelClickHandler() {
    this._CancelRetry();
    this.UploadItem.CancelAsync();
  }
  _PauseClickHandler() {
    this._CancelRetry();
    this.UploadItem.PauseAsync();
  }
  _PlayClickHandler() {
    this._CurrentRetry = 0;
    this.UploadItem.StartAsync();
  }

  _SetRetryMessage(timeLeft) {
    var sRetryMessageFormat = "Retry in: {0}";
    this.RetryMessage = CommonService.Validators.PasteFormat(
      sRetryMessageFormat,
      CommonService.Formatters.TimeSpan(Math.ceil(timeLeft / 1000))
    );
  }
  _RemoveRetryMessage() {
    this.RetryMessage = null;
  }
  _OnUploadError = oUploadError => {
    // Here you can verify error code returned by the server and show error UI,
    // for example if server-side validation failed.

    // Stop upload if max upload retries reached.
    if (this._MaxRetry <= this._CurrentRetry) {
      //this._ShowError(oUploadError.Error);
      oUploadError.Skip();
      return;
    }
    var me = this;
    // Retry upload.
    var retryTime = new Date().getTime() + this._RetryDelay * 1000;
    var retryTimerId = setInterval(function() {
      var timeLeft = retryTime - new Date().getTime();
      if (timeLeft > 0) {
        me._SetRetryMessage(timeLeft);
        return;
      }
      clearInterval(retryTimerId);
      me._CurrentRetry++;
      me._RemoveRetryMessage();

      // Request number of bytes succesefully saved on the server
      // and retry upload from next byte.
      oUploadError.Retry();
    }, 1000);
    this.CancelRetryCallback = function() {
      clearInterval(retryTimerId);
      this._RemoveRetryMessage();
    };
  };

  _CancelRetry() {
    if (this.CancelRetryCallback) this.CancelRetryCallback.call(this);
  }
}
