import moment from "moment";
import i18n from "@/i18n";

const sFileNameSpecialCharactersRestrictionFormat =
  i18n.t("phrases.validations.notContainFollowingCharacters") + ": {0}";
const sForbiddenNameChars = '\\/:*?"<>|';
export default {
  Formatters: {
    /**
     * Gets file extension
     * @param {string} fileName file name with extension
     * @returns {string} file extension
     */
    GetFileExtension: function(fileName) {
      var index = fileName.lastIndexOf(".");
      return index !== -1 ? fileName.substr(index + 1).toLowerCase() : "";
    },
    /**
     * Formats file size
     * @param {number} iSize file size
     * @returns {string} formated file size
     */
    FileSize: function(iSize) {
      if (!iSize) {
        return "0.00 B";
      }
      var i = Math.floor(Math.log(iSize) / Math.log(1024));
      return (
        (iSize / Math.pow(1024, i)).toFixed(2) +
        " " +
        ["B", "kB", "MB", "GB", "TB"][i]
      );
    },
    /**
     * Formats date
     * @param {Date} oDate date
     * @returns {string} formated date
     */
    Date: function(oDate) {
      return moment(oDate).fromNow();
    },
    /**
     * Formats snippet
     * @param {string} html
     * @returns {string}
     */
    Snippet: function(html) {
      if (html) {
        var safePrefix = "__b__tag" + new Date().getTime();
        html = html
          .replace(/<b>/g, safePrefix + "_0")
          .replace(/<\/b>/g, safePrefix + "_1");
        html = "<div>" + html + "</div>";
        html = html
          .replace(new RegExp(safePrefix + "_0", "g"), "<b>")
          .replace(new RegExp(safePrefix + "_1", "g"), "</b>");
      }
      return html;
    },
    /**
     * Converts a string to an HTML-encoded string.
     * @param {String} sText - The string to encode.
     * @return {String} - An encoded string.
     */
    HtmlEscape: function(sText) {
      return String(sText)
        .replace(/&/g, "&amp;")
        .replace(/"/g, "&quot;")
        .replace(/'/g, "&#39;")
        .replace(/</g, "&lt;")
        .replace(/>/g, "&gt;");
    },
    /**
     *
     * @param {number} iSeconds
     * @returns {string}
     */
    TimeSpan: function(iSeconds) {
      var hours = Math.floor(iSeconds / 3600);
      var minutes = Math.floor((iSeconds - hours * 3600) / 60);
      var seconds = iSeconds - hours * 3600 - minutes * 60;
      var sResult = "";
      if (hours) sResult += hours + "h ";
      if (minutes) sResult += minutes + "m ";
      sResult += seconds + "s ";
      return sResult;
    }
  },
  Validators: {
    /**
     * Validateds name
     * @param {String} sName
     */
    ValidateName: function(sName) {
      var oRegExp = new RegExp("[" + sForbiddenNameChars + "]", "g");
      if (oRegExp.test(sName)) {
        var sMessage = this.PasteFormat(
          sFileNameSpecialCharactersRestrictionFormat,
          sForbiddenNameChars.replace(/\\?(.)/g, "$1 ")
        );
        return sMessage;
      }
    },
    PasteFormat: function(sPhrase) {
      var callbackReplace = function(oArguments) {
        this._arguments = oArguments;
      };

      callbackReplace.prototype.Replace = function(sPlaceholder) {
        var iIndex = sPlaceholder.substr(1, sPlaceholder.length - 2);
        return "undefined" !== typeof this._arguments[iIndex]
          ? this._arguments[iIndex]
          : sPlaceholder;
      };

      if (/\{\d+?\}/.test(sPhrase)) {
        var oReplace = new callbackReplace(
          Array.prototype.slice.call(arguments, 1)
        );
        sPhrase = sPhrase.replace(/\{(\d+?)\}/g, function(args) {
          return oReplace.Replace(args);
        });
      }

      return sPhrase;
    }
  },

  ToolbarHelpers: {
    /**
     * Gets name with copy suffix or number of copies
     * @param {String} oItemName Name of HierachyItem
     */
    GetCopySuffix: function(oItemName, bWithCopySuffix) {
      var sCopySuffixName = "Copy";

      var aExtensionMatches = /\.[^\\.]+$/.exec(oItemName);
      var sName =
        aExtensionMatches !== null
          ? oItemName.replace(aExtensionMatches[0], "")
          : oItemName;
      var sDotAndExtension =
        aExtensionMatches !== null ? aExtensionMatches[0] : "";

      var sLangCopy = sCopySuffixName;
      var oSuffixPattern = new RegExp(
        "- " + sLangCopy + "( \\(([0-9]+)\\))?$",
        "i"
      );

      var aSuffixMatches = oSuffixPattern.exec(sName);
      if (aSuffixMatches === null && bWithCopySuffix) {
        sName += " - " + sLangCopy;
      } else if (aSuffixMatches !== null && !aSuffixMatches[1]) {
        sName += " (2)";
      } else if (aSuffixMatches !== null) {
        var iNextNumber = parseInt(aSuffixMatches[2]) + 1;
        sName = sName.replace(
          oSuffixPattern,
          "- " + sLangCopy + " (" + iNextNumber + ")"
        );
      }

      oItemName = sName + sDotAndExtension;
      return oItemName;
    }
  }
};
