import { WebDavStore } from "@/stores/WebDavStore.js";
import { UploadItemRow } from "@/models/UploadItemRow";
import CommonService from "@/services/CommonService.js";
import UrlResolveService from "@/services/UrlResolveService.js";
import { WebDavError } from "@/models/WebDavError.js";
import { RewriteItemsData } from "@/models/RewriteItemsData.js";
import WebDavService from "@/services/WebDavService.js";
import { ITHit } from "webdav.client";
import i18n from "@/i18n";
export default {
  Uploader: null,
  init() {
    this.Uploader = new ITHit.WebDAV.Client.Upload.Uploader();

    var path = UrlResolveService.getWebDavServerUrl(
      WebDavStore.state.grid.CurrentUrl
    );

    this.Uploader.Inputs.AddById("ithit-hidden-input");
    this.Uploader.Inputs.AddById("ithit-button-input");
    this.Uploader.DropZones.AddById("ithit-dropzone");

    this.Uploader.SetUploadUrl(ITHit.WebDAV.Client.Encoder.Decode(path));
    this.Uploader.Queue.AddListener("OnQueueChanged", this._queueChange, this);
    this.Uploader.Queue.AddListener(
      "OnUploadItemsCreated",
      this._onUploadItemsCreated,
      this
    );
  },

  _queueChange: function(oQueueChanged) {
    // Display each item added to the upload queue in the grid.
    oQueueChanged.AddedItems.forEach(
      function(value) {
        var row = new UploadItemRow(value);
        WebDavStore.commit("addUploadingItem", row);
      }.bind(this)
    );

    oQueueChanged.RemovedItems.forEach(
      function(value) {
        WebDavStore.commit("removeUploadingItem", value);
      }.bind(this)
    );
  },

  _onUploadItemsCreated: function(oUploadItemsCreated) {
    /* Validate file extensions, size, name, etc. here. */
    var oValidationError = this._validateUploadItems(oUploadItemsCreated.Items);
    if (oValidationError) {
      WebDavStore.commit(
        "setError",
        new WebDavError(
          i18n.t("phrases.errors.validationErrorMessage"),
          oValidationError
        )
      );
      return;
    }

    /* Below we will check if each file exists on the server 
        and ask a user if files should be overwritten or skipped. */
    this._getExistsAsync(oUploadItemsCreated.Items, function(oAsyncResult) {
      if (oAsyncResult.IsSuccess && oAsyncResult.Result.length === 0) {
        // No items exists on the server.
        // Add all items to the upload queue.
        oUploadItemsCreated.Upload(oUploadItemsCreated.Items);
        return;
      }
      if (!oAsyncResult.IsSuccess) {
        // Some error occurred during item existence verification requests.
        // Show error dialog with error description.
        // Mark all items as failed and add to the upload list.
        WebDavStore.commit(
          "setError",
          new WebDavError(
            i18n.t("phrases.errors.failedCheckExistsErrorMessage"),
            oAsyncResult.Error
          )
        );

        oUploadItemsCreated.Items.forEach(function(oUploadItem) {
          // Move an item into the error state.
          // Upload of this item will NOT start when added to the queue.
          oUploadItem.SetFailed(oAsyncResult.Error);
        });

        // Add all items to the upload queue, so a user can start the upload later.
        oUploadItemsCreated.Upload(oUploadItemsCreated.Items);

        return;
      }

      var sItemsList = ""; // List of items to be displayed in Overwrite / Skip / Cancel dialog.

      /** @type {ITHit.WebDAV.Client.Upload.UploadItem[]} aExistsUploadItems */
      var aExistsUploadItems = [];
      oAsyncResult.Result.forEach(function(oUploadItem) {
        // For the sake of simplicity folders are never deleted when upload canceled.
        if (!oUploadItem.IsFolder()) {
          // File exists so we should not delete it when file's upload canceled.
          oUploadItem.SetDeleteOnCancel(false);
        }

        // Mark item as verified to avoid additional file existence verification requests.
        oUploadItem.CustomData.FileExistanceVerified = true;

        sItemsList += oUploadItem.GetRelativePath() + "<br/>";
        aExistsUploadItems.push(oUploadItem);
      });
      var onOverwrite = function() {
        // Mark all items that exist on the server with overwrite flag.
        aExistsUploadItems.forEach(function(oUploadItem) {
          if (oUploadItem.IsFolder()) return;

          // The file will be overwritten if it exists on the server.
          oUploadItem.SetOverwrite(true);
        });

        // Add all items to the upload queue.
        oUploadItemsCreated.Upload(oUploadItemsCreated.Items);
      };

      var onSkipExists = function() {
        // Create list of items that do not exist on the server.
        /** @type {ITHit.WebDAV.Client.Upload.UploadItem[]} aNotExistsUploadItems */
        var grep = function(elems, callback, invert) {
          var callbackInverse,
            matches = [],
            i = 0,
            length = elems.length,
            callbackExpect = !invert;

          // Go through the array, only saving the items
          // that pass the validator function
          for (; i < length; i++) {
            callbackInverse = !callback(elems[i], i);
            if (callbackInverse !== callbackExpect) {
              matches.push(elems[i]);
            }
          }

          return matches;
        };
        var aNotExistsUploadItems = grep(oUploadItemsCreated.Items, function(
          oUploadItem
        ) {
          return !ITHit.Utils.Contains(aExistsUploadItems, oUploadItem);
        });

        // Add only items that do not exist on the server to the upload queue.
        oUploadItemsCreated.Upload(aNotExistsUploadItems);
      };
      /* One or more items exists on the server. Show Overwrite / Skip / Cancel dialog.*/
      var rewriteData = new RewriteItemsData(
        /* A user selected to overwrite existing files. */
        onOverwrite,
        /* A user selected to skip existing files. */
        onSkipExists,
        sItemsList
      );
      WebDavStore.commit("setRewriteItemsData", rewriteData);
    });
  },

  _validateUploadItems(aUploadItems) {
    for (var i = 0; i < aUploadItems.length; i++) {
      var oUploadItem = aUploadItems[i];
      //Max file size validation
      //var oExtensionError = this._validateExtension(oUploadItem);

      //File extension validation
      //var oSizeError = this._validateSize(oUploadItem);

      //Special characters validation
      //var oNameError = this._validateName(oUploadItem);

      //var oValidationError = oExtensionError || oSizeError || oNameError;
      //if(oValidationError) {
      //    return oValidationError;
      //}

      var oValidationError = this._validateName(oUploadItem);
      if (oValidationError) {
        return oValidationError;
      }
    }
  },

  _validateName(oUploadItem) {
    var sValidationMessage = CommonService.Validators.ValidateName(
      oUploadItem.GetName()
    );
    if (sValidationMessage) {
      return new { Message: sValidationMessage, Uri: oUploadItem.GetUrl() }();
    }
  },

  // _validateExtension(oUploadItem) {
  //   var aForbiddenExtensions = ["BAT", "BIN", "CMD", "COM", "EXE"];
  //   var sForbiddenExtensionFormat =
  //     'Upload files with "{0}" extension is forbidden.';
  //   var sExtension = CommonService.Formatters.GetExtension(
  //     oUploadItem.GetUrl()
  //   );

  //   if (aForbiddenExtensions.indexOf(sExtension.toUpperCase()) >= 0) {
  //     var sMessage = CommonService.PasteFormat(
  //       sForbiddenExtensionFormat,
  //       sExtension
  //     );
  //     return new { Message: sMessage, Uri: oUploadItem.GetUrl() }();
  //   }
  // },

  // _validateSize(oUploadItem) {
  //   var iMaxFileSize = 10485760; //10MB
  //   var sWrongFileSizeFormat = "File size should be less than {0}.";
  //   if (oUploadItem.GetSize() > iMaxFileSize) {
  //     var sMessage = CommonService.PasteFormat(
  //       sWrongFileSizeFormat,
  //       CommonService.Formatters.FileSize(iMaxFileSize)
  //     );
  //     return new { Message: sMessage, Uri: oUploadItem.GetUrl() }();
  //   }
  // },

  _getExistsAsync: function(aUploadItems, fCallback) {
    this._openItemsCollectionAsync(aUploadItems, function(aResultCollection) {
      var oFailedResult = ITHit.Utils.FindBy(
        aResultCollection,
        function(oResult) {
          return !(
            oResult.AsyncResult.IsSuccess ||
            oResult.AsyncResult.Status.Code === 404
          );
        },
        this
      );

      if (oFailedResult) {
        fCallback(oFailedResult.AsyncResult);
        return;
      }

      var aExistsItems = aResultCollection
        .filter(function(oResult) {
          return oResult.AsyncResult.IsSuccess;
        })
        .map(function(oResult) {
          return oResult.UploadItem;
        });
      fCallback(new ITHit.WebDAV.Client.AsyncResult(aExistsItems, true, null));
    });
  },

  _openItemsCollectionAsync: function(aUploadItems, fCallback) {
    var iCounter = aUploadItems.length;
    /**@type {UploaderGridView~OpenItemsCollectionResult} */
    var aResults = [];
    if (iCounter === 0) {
      fCallback(aResults);
      return;
    }

    aUploadItems.forEach(function(oUploadItem) {
      WebDavService.WebDavSession.OpenItemAsync(
        ITHit.EncodeURI(oUploadItem.GetUrl()),
        [],
        function(oAsyncResult) {
          iCounter--;
          aResults.push({
            UploadItem: oUploadItem,
            AsyncResult: oAsyncResult
          });

          if (iCounter === 0) {
            fCallback(aResults);
          }
        }
      );
    });
  }
};
