import { WebDavStore } from "@/stores/WebDavStore.js";
import webDavSettings from "@/WebDavSettings.js";

export default {
  /**
   * Gets webdav server url
   * @param {String} sUrl Relative path
   */
  getWebDavServerUrl(sUrl = "") {
    let serverUrl;
    if (!webDavSettings.WebDavServerPath) {
      serverUrl = window.location.origin + sUrl;
    } else {
      serverUrl = new URL(webDavSettings.WebDavServerPath);
    }

    if (sUrl) {
      return serverUrl + this.getTail(serverUrl.origin + sUrl, serverUrl);
    }
    return serverUrl;
  },
  /**
   * Gets applicataion url
   * @param {String} sUrl Webdav server path
   * @param {Boolean} clearQuery
   */
  getApplicationUrl(sUrl, clearQuery = false) {
    var appUrl = sUrl;
    var oUrl = new URL(sUrl);

    if (oUrl.origin !== window.location.origin) {
      appUrl = sUrl.replace(
        webDavSettings.WebDavServerPath,
        window.location.origin
      );
    }

    if (clearQuery) {
      oUrl = new URL(appUrl);
      appUrl = oUrl.origin + oUrl.pathname;
    }
    return appUrl;
  },
  /**
   * Sets page number to store
   * @param {Number} pageNumber Page Number
   */
  setPage(pageNumber) {
    WebDavStore.commit("setCurrentPage", pageNumber);
    return { path: WebDavStore.state.grid.CurrentUrl, hash: this._getHash() };
  },
  /**
   * Gets relative url
   * @param {String} sUrl
   */
  getRelativeUrl(sUrl) {
    var tempUrl = new URL(sUrl);
    return { path: tempUrl.pathname };
  },
  /**
   * Sets search query to store
   * @param {String} query Search query
   */
  setSearch(query) {
    WebDavStore.commit("setSearchQuery", query);
    return { path: WebDavStore.state.grid.CurrentUrl, hash: this._getHash() };
  },
  /**
   * Sets sort params query to store
   * @param {String} sortColumn
   * @param {Boolean} sortAscending
   */
  setSortColumn(sortColumn, sortAscending) {
    WebDavStore.commit("setSortColumn", sortColumn);
    WebDavStore.commit("setSortAscending", sortAscending);
    return { path: WebDavStore.state.grid.CurrentUrl, hash: this._getHash() };
  },
  /**
   * Sets hash to store
   * @param {String} sHash
   */
  setHashObject(sHash) {
    var pageNumber = this._getHashValue("page", sHash);
    var searchQuery = this._getHashValue("search", sHash);
    var sortColumn = this._getHashValue("sortcolumn", sHash);
    var sortAscending = this._getHashValue("sortascending", sHash);
    WebDavStore.commit("setCurrentPage", pageNumber == null ? 1 : pageNumber);
    WebDavStore.commit(
      "setSearchQuery",
      searchQuery == null ? "" : searchQuery
    );
    WebDavStore.commit("setSortColumn", sortColumn == null ? "" : sortColumn);
    WebDavStore.commit(
      "setSortAscending",
      sortAscending == null ? true : sortAscending == "true"
    );
  },

  getTail(url1, url2) {
    return url1.includes(url2) ? url1.replace(url2, "") : "";
  },

  _getHashValue(key, sHash) {
    var hashConfig = this._parseUrlHash(sHash);

    return Object.prototype.hasOwnProperty.call(hashConfig, key)
      ? hashConfig[key]
      : null;
  },

  _parseUrlHash: function(sHash) {
    // Parse hash
    var hash = {};
    if (sHash.length > 0) {
      var hashParts = sHash.substr(1).split("&");
      for (var i = 0, l = hashParts.length; i < l; i++) {
        var param = hashParts[i].split("=");
        hash[param[0]] = param[1];
      }
    }

    return hash;
  },

  _getHash: function() {
    var hash = [];
    if (WebDavStore.state.grid.SearchQuery) {
      hash.push("search=" + WebDavStore.state.grid.SearchQuery);
    }
    if (WebDavStore.state.grid.CurrentPage > 1) {
      hash.push("page=" + WebDavStore.state.grid.CurrentPage);
    }
    if (WebDavStore.state.grid.SortColumn) {
      hash.push("sortcolumn=" + WebDavStore.state.grid.SortColumn);
      hash.push("sortascending=" + WebDavStore.state.grid.SortAscending);
    }
    return hash.length ? "#" + hash.join("&") : "";
  }
};
