import { HierarchyItem } from "@/models/HierarchyItem.js";
import { ITHit } from "webdav.client";

const webDavSession = new ITHit.WebDAV.Client.WebDavSession();
const searchQuery = new ITHit.WebDAV.Client.SearchQuery();
const snippetPropertyName = new ITHit.WebDAV.Client.PropertyName(
  "snippet",
  "ithit"
);
export default {
  SnippetPropertyName: snippetPropertyName,
  DocManager: ITHit.WebDAV.Client.DocManager,
  WebDavSession: webDavSession,
  /**
   * Gets current folder
   * @param {String} sPath folder path
   */
  getCurrentFolder: function(sPath) {
    return new Promise((resolve, reject) => {
      webDavSession.OpenFolderAsync(sPath, [], data => {
        if (!data.IsSuccess) return reject(data);
        resolve(data);
      });
    });
  },
  /**
   * Gets item
   * @param {String} sPath item path
   */
  getItem: function(sPath) {
    return new Promise((resolve, reject) => {
      webDavSession.OpenItemAsync(sPath, [], data => {
        if (!data.IsSuccess) return reject(data);
        resolve(data);
      });
    });
  },

  /**
   * Gets supported features
   * @param {ITHit.WebDAV.Client.Folder} currentFolder Current folder
   */
  getSupportedFeatures: function(currentFolder) {
    return new Promise((resolve, reject) => {
      currentFolder.GetSupportedFeaturesAsync(data => {
        if (!data.IsSuccess) return reject(data);
        resolve(data);
      });
    });
  },
  /**
   * Gets all items in current folder
   * @param {ITHit.WebDAV.Client.Folder} currentFolder Current folder
   * @param {String} sortColumn name of the sorted column
   * @param {Boolean} sortAscending sort ascending or descending
   * @param {Number} currentPage current page
   * @param {Number} pageSize Size of page
   */
  getItems: function(
    currentFolder,
    sortColumn,
    sortAscending,
    currentPage,
    pageSize
  ) {
    return new Promise((resolve, reject) => {
      currentFolder.GetPageAsync(
        [],
        (currentPage - 1) * pageSize,
        pageSize,
        this._setSortColumnsValue(sortColumn, sortAscending),
        data => {
          if (!data.IsSuccess) return reject(data);
          resolve(this._convertItems(data));
        }
      );
    });
  },
  /**
   * Gets items by query in current folder
   * @param {ITHit.WebDAV.Client.Folder} currentFolder Current folder
   * @param {Number} currentPage current page
   * @param {Number} pageSize Size of page
   * @param {String} sPhrase Search query
   */
  getItemsByQuery: function(currentFolder, currentPage, pageSize, sPhrase) {
    searchQuery.Phrase =
      sPhrase
        .replace(/\\/g, "\\\\")
        .replace(/\\%/g, "\\%")
        .replace(/\\_/g, "\\_")
        .replace(/\*/g, "%")
        .replace(/\?/g, "_") + "%";
    searchQuery.SelectProperties = [
      new ITHit.WebDAV.Client.PropertyName("snippet", "ithit")
    ];
    return new Promise((resolve, reject) => {
      currentFolder.GetSearchPageByQueryAsync(
        searchQuery,
        (currentPage - 1) * pageSize,
        pageSize,
        data => {
          if (!data.IsSuccess) return reject(data);
          resolve(this._convertItems(data));
        }
      );
    });
  },
  /**
   * Cteate folder
   * @param {ITHit.WebDAV.Client.Folder} currentFolder Current folder
   * @param {String} folderName Name of folder
   */
  createFolder: function(currentFolder, folderName) {
    return new Promise((resolve, reject) => {
      currentFolder.CreateFolderAsync(folderName, null, null, data => {
        if (!data.IsSuccess) return reject(data);
        resolve(data);
      });
    });
  },

  _convertItems: function(resp) {
    var hierarchyItems = [];
    resp.Result.Page.forEach(item => {
      hierarchyItems.push(new HierarchyItem(item));
    });
    return {
      Page: hierarchyItems,
      TotalItems: resp.Result.TotalItems
    };
  },

  _setSortColumnsValue: function(sortColumn, sortAscending) {
    var sortColumns = [
      new ITHit.WebDAV.Client.OrderProperty(
        new ITHit.WebDAV.Client.PropertyName(
          "is-directory",
          ITHit.WebDAV.Client.DavConstants.NamespaceUri
        ),
        ITHit.WebDAV.Client.OrderProperty.CurrentSortColumnAscending
      )
    ];
    if (sortColumn) {
      sortColumns.push(
        new ITHit.WebDAV.Client.OrderProperty(
          new ITHit.WebDAV.Client.PropertyName(
            sortColumn,
            ITHit.WebDAV.Client.DavConstants.NamespaceUri
          ),
          sortAscending
        )
      );
    }
    return sortColumns;
  }
};
