import Vue from "vue";
import webDavSettings from "@/WebDavSettings.js";
import WebDavService from "@/services/WebDavService.js";
import CommonService from "@/services/CommonService.js";
import UrlResolveService from "@/services/UrlResolveService.js";
import { StoredType } from "@/models/StoredType.js";
import { CopyHierarchyItem } from "@/models/CopyHierarchyItem.js";
import { WebDavError } from "@/models/WebDavError.js";
import { HierarchyItem } from "@/models/HierarchyItem.js";
import { ITHit } from "webdav.client";
import i18n from "@/i18n";
export default {
  state: {
    Loading: true,
    GridLoading: true,
    Items: [],
    SelectedItems: [],
    //Copied or cut items
    StoredItems: [],
    StoredType: StoredType.Copy,
    CurrentUrl: "",
    CurrentFolder: null,
    CurrentPage: 1,
    PageSize: 10,
    CountPages: 0,
    SearchQuery: "",
    Error: null,
    ShowProtocolModal: false,
    SortColumn: "",
    SortAscending: true,
    OptionsInfoLoading: true,
    OptionsInfo: null
  },
  mutations: {
    /**
     * Gets all or searched items in current folder
     * @param {ITHit.WebDAV.Client.Folder} currnetFolder
     */
    getItems(state, currnetFolder) {
      var getItems = state.SearchQuery
        ? WebDavService.getItemsByQuery(
            currnetFolder,
            state.CurrentPage,
            state.PageSize,
            state.SearchQuery
          )
        : WebDavService.getItems(
            currnetFolder,
            state.SortColumn,
            state.SortAscending,
            state.CurrentPage,
            state.PageSize
          );
      getItems
        .then(resp => {
          state.Items = resp.Page;
          state.CountPages = Math.ceil(resp.TotalItems / state.PageSize);
          state.Loading = false;
        })
        .catch(resp => {
          this.commit(
            "setError",
            new WebDavError(
              i18n.t("phrases.errors.profindErrorMessage"),
              resp.Error
            )
          );
        });
    },
    /**
     * Updates grid view
     */
    updateGrid(state) {
      state.GridLoading = true;
      WebDavService.getCurrentFolder(
        UrlResolveService.getWebDavServerUrl(state.CurrentUrl)
      )
        .then(resp => {
          if (resp.Result) {
            state.CurrentFolder = resp.Result;
            this.commit("getItems", state.CurrentFolder);
            this.commit("getSupportedFeatures");
          }
        })
        .catch(resp => {
          this.commit(
            "setError",
            new WebDavError(
              i18n.t("phrases.errors.profindErrorMessage"),
              resp.Error
            )
          );
        });
    },
    /**
     * Updates item
     */
    updateItem(state, itemPath) {
      var item = state.Items.find(p => p.Href == itemPath);
      if (item) {
        WebDavService.getItem(item.Href)
          .then(resp => {
            if (resp.Result) {
              let index = state.Items.findIndex(p => p.Href == itemPath);
              if (index >= 0) {
                Vue.set(state.Items, index, new HierarchyItem(resp.Result));
              }
            }
          })
          .catch(resp => {
            this.commit(
              "setError",
              new WebDavError(
                i18n.t("phrases.errors.profindErrorMessage"),
                resp.Error
              )
            );
          });
      }
    },
    /**
     * Gets supported features
     */
    getSupportedFeatures(state) {
      WebDavService.getSupportedFeatures(state.CurrentFolder)
        .then(resp => {
          state.OptionsInfo = resp.Result;
          state.OptionsInfoLoading = false;
        })
        .catch(resp => {
          state.OptionsInfoLoading = false;
          this.commit(
            "setError",
            new WebDavError(
              i18n.t("phrases.errors.supportedFeaturesErrorMessage"),
              resp.Error
            )
          );
        });
    },
    /**
     * Sets GridLoading state
     * @param {Boolean} loadingState
     */
    setGridLoading(state, loadingState) {
      state.GridLoading = loadingState;
    },
    /**
     * Sets searched item state
     * @param {HierarchyItem} item
     */
    setSearchItem(state, item) {
      var arr = [];
      arr.push(item);
      state.Items = arr;
      this.commit("setCurrentPage", 1);
      state.CountPages = 1;
    },
    /**
     * Sets search query state
     * @param {String} searchQuery
     */
    setSearchQuery(state, searchQuery) {
      state.SearchQuery = searchQuery;
    },
    /**
     * Sets page state
     * @param {Number} pageNumber
     */
    setCurrentPage(state, pageNumber) {
      state.CurrentPage = pageNumber;
    },
    /**
     * Sets url state
     * @param {String} sUrl
     */
    setCurrentUrl(state, sUrl) {
      state.CurrentUrl = sUrl;
    },
    /**
     * Adds item to SelectedItems array
     * @param {HierarchyItem} item
     */
    addSelectedItem(state, item) {
      state.SelectedItems.push(item);
    },
    /**
     * Removes item from SelectedItems array
     * @param {HierarchyItem} item
     */
    removeSelectedItem(state, item) {
      state.SelectedItems.forEach((selectedItem, index) => {
        if (selectedItem.Href === item.Href) {
          state.SelectedItems.splice(index, 1);
          return false;
        }
      });
    },
    /**
     * Clears SelectedItems array
     */
    clearSelectedItems(state) {
      state.SelectedItems = [];
    },
    /**
     * Clones SelectedItems to StoreItems
     * @param {StoredType} type
     */
    storeSelectedItems(state, type) {
      state.StoredItems = [...state.SelectedItems];
      state.StoredType = type;
    },
    /**
     * Pastes StoredItems array
     */
    pasteStoredItems(state) {
      switch (state.StoredType) {
        case StoredType.Cut:
          this.commit("moveStoredItems");
          break;
        case StoredType.Copy:
          this.commit("copyStoredItems");
          break;
      }
    },
    /**
     * Moves StoredItems array
     */
    moveStoredItems(state) {
      state.StoredItems.forEach(item => {
        item
          .MoveToAsync(state.CurrentFolder)
          .then(() => {
            state.StoredItems = [];
            this.commit("updateGrid");
          })
          .catch(resp => {
            var error;
            if (
              resp.Error instanceof
              ITHit.WebDAV.Client.Exceptions.ForbiddenException
            ) {
              error = new WebDavError(
                i18n.t("phrases.errors.cutItemsSameNameErrorMessage"),
                resp.Error
              );
            } else if (
              resp.Error instanceof
              ITHit.WebDAV.Client.Exceptions.LockedException
            ) {
              error = new WebDavError(
                i18n.t("phrases.errors.cutItemsLockedErrorMessage"),
                resp.Error
              );
            } else {
              error = new WebDavError(
                i18n.t("phrases.errors.cutItemsErrorMessage"),
                resp.Error
              );
            }
            this.commit("setError", error);
          });
      });
    },
    /**
     * Renames the first element of the StoredItems array
     */
    renameSelectedItem(state, itemName) {
      if (state.SelectedItems.length == 1) {
        state.SelectedItems[0]
          .MoveToAsync(state.CurrentFolder, itemName)
          .then(() => {
            this.commit("updateGrid");
          })
          .catch(resp => {
            var error;
            if (
              resp.Error instanceof
              ITHit.WebDAV.Client.Exceptions.LockedException
            ) {
              error = new WebDavError(
                i18n.t("phrases.errors.renameItemLockedErrorMessage"),
                resp.Error
              );
            } else {
              error = new WebDavError(
                i18n.t("phrases.errors.renameItemErrorMessage"),
                resp.Error
              );
            }
            this.commit("setError", error);
          });
      }
    },
    /**
     * Deletes items from SelectedItems array
     */
    deleteSelectedItems(state) {
      state.SelectedItems.forEach(item => {
        item
          .DeleteAsync()
          .then(() => {
            this.commit("updateGrid");
          })
          .catch(resp => {
            this.commit(
              "setError",
              new WebDavError(
                i18n.t("phrases.errors.deletetemsErrorMessage"),
                resp.Error
              )
            );
          });
      });
    },
    /**
     * Copies StoredItems array
     */
    copyStoredItems(state) {
      state.StoredItems.forEach(item => {
        var itemCopyName = CommonService.ToolbarHelpers.GetCopySuffix(
          item.DisplayName,
          false
        );
        this.commit(
          "copyStoredItem",
          new CopyHierarchyItem(item, itemCopyName)
        );
      });
    },
    /**
     * Copies stored Item
     * @param {CopyHierarchyItem} copyItem
     */
    copyStoredItem(state, copyItem) {
      copyItem.HierarchyItem.CopyToAsync(
        state.CurrentFolder,
        copyItem.ItemCopyName
      )
        .then(() => {
          this.commit("updateGrid");
        })
        .catch(resp => {
          if (
            resp.Error instanceof
              ITHit.WebDAV.Client.Exceptions.PreconditionFailedException ||
            resp.Error instanceof
              ITHit.WebDAV.Client.Exceptions.ForbiddenException
          ) {
            copyItem.ItemCopyName = CommonService.ToolbarHelpers.GetCopySuffix(
              copyItem.ItemCopyName,
              true
            );
            this.commit("copyStoredItem", copyItem);
          } else {
            this.commit(
              "setError",
              new WebDavError(
                i18n.t("phrases.errors.copyItemsErrorMessage"),
                resp.Error
              )
            );
          }
        });
    },
    /**
     * Prints documents
     */
    printDocs(state) {
      var filesUrls = [];
      state.SelectedItems.forEach(item => {
        if (!item.IsFolder) {
          filesUrls.push(item.Href);
        }
      });

      WebDavService.DocManager.DavProtocolEditDocument(
        filesUrls,
        UrlResolveService.getWebDavServerUrl(),
        () => {
          this.commit("showProtocolModal");
        },
        null,
        webDavSettings.EditDocAuth.SearchIn,
        webDavSettings.EditDocAuth.CookieNames,
        webDavSettings.EditDocAuth.LoginUrl,
        "Print"
      );
    },
    /**
     * Sets state Error
     * @param {WebDavError} webDavError
     */
    setError(state, webDavError) {
      state.Error = webDavError;
    },
    /**
     * Clears state Error
     */
    clearError(state) {
      state.Error = null;
    },
    /**
     * Shows Protocol modal
     */
    showProtocolModal(state) {
      if (ITHit.WebDAV.Client.DocManager.IsDavProtocolSupported()) {
        state.ShowProtocolModal = true;
      }
    },
    /**
     * Hides Protocol modal
     */
    hideProtocolModal(state) {
      state.ShowProtocolModal = false;
    },
    /**
     * Sets SortColumn state
     * @param {String} sortColumn
     */
    setSortColumn(state, sortColumn) {
      state.SortColumn = sortColumn;
    },
    /**
     * Sets SortAscending state
     * @param {Boolean} sortAscending
     */
    setSortAscending(state, sortAscending) {
      state.SortAscending = sortAscending;
    }
  }
};
