//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Hosting != HostingEnvironment.IIS)  return string.Empty; #>
//!<# if (!Params.AuthenticationType.HasFlag(AuthenticationType.Basic))  return string.Empty; #>
//!<# if (!Params.IsOwinAuthMode) return string.Empty; #>
using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Claims;
using System.Text;
using System.Threading.Tasks;
using Microsoft.AspNet.Identity;
using Microsoft.AspNet.Identity.Owin;
using Microsoft.Owin;
using Microsoft.Owin.Security;
using Microsoft.Owin.Security.Infrastructure;

namespace ITHit.WebDav.Server.SampleModel.__Authentication__.BasicAuthentication
{
    /// <summary>
    /// OWIN middleware that implements 'Basic' authentication protocol.
    /// </summary>
    /// <remarks>
    /// Provides fallback to Basic authentication if HTTP response code 302 
    /// (redirect to log-in page) has been issued after (failed) attempt to 
    /// authenticate using other OWIN authentication middleware (i.e. cookie)
    /// that isn't supported by host (i.e. Microsoft Office application).
    /// </remarks>
    internal class BasicAuthenticationHandler<TKey, TUser, TManager>
        : AuthenticationHandler<BasicAuthenticationOptions<TKey, TUser, TManager>>
        where TKey : IEquatable<TKey>
        where TUser : class, IUser<TKey>
        where TManager : UserManager<TUser, TKey>
    {
        private const string defaultUnauthorizedResponseHtml =
            @"<html><title>401 Authorization Required</title></html>";

        private readonly string challenge;

        public BasicAuthenticationHandler(BasicAuthenticationOptions<TKey, TUser, TManager> options)
        {
            challenge = string.Format("Basic realm=\"{0}\"", options.Realm);
        }

        /// <summary>
        /// Validates Basic authentication credentials using the same algorithm
        /// as used by Cookie authentication middleware.
        /// </summary>
        /// <typeparam name="TKey">Type of the primary key used by ASP.NET MVC / Entity Framework.
        /// string type is used by default, unless you've configured your project to use int primary keys.</typeparam>
        /// <typeparam name="TUser">Application user class (typically ApplicationUser).</typeparam>
        /// <typeparam name="TManager">Application user manager class (typically ApplicationUserManager).</typeparam>
        /// <param name="id">Username to be validated. Note that default ASP.NET MVC project template 
        /// uses e-mail as username during log-in / registration, so the user may typically enter their 
        /// e-mail address when prompted during Basic authentication.
        /// </param>
        /// <param name="secret">Password to be validated. 
        /// Same as the user's password registered in ASP.NET MVC application</param>
        /// <param name="context">Class that wraps OWIN environment and provides 
        /// access to core functionality of the OWIN middleware.</param>
        /// <returns>Enumeration of claims associated with created user identity 
        /// (if successful), or null if log-in has failed.</returns>
        public static async Task<IEnumerable<Claim>> ValidateUser(string id, string secret, IOwinContext context)
        {
            TManager manager = context.GetUserManager<TManager>();
            TUser user = await manager.FindAsync(id, secret);

            if (user != null)
            {
                ClaimsIdentityFactory<TUser, TKey> factory = new ClaimsIdentityFactory<TUser, TKey>();
                ClaimsIdentity identity = await factory.CreateAsync(manager, user,
                    DefaultAuthenticationTypes.ApplicationCookie);

                if (identity != null)
                    return identity.Claims;
            }

            return null;
        }

        /// <summary>
        /// Performs request authentication.
        /// </summary>
        /// <returns>
        /// <see cref="AuthenticationTicket"/> class that contains user identity information 
        /// as well as additional authentication state.
        /// </returns>
        protected override async Task<AuthenticationTicket> AuthenticateCoreAsync()
        {
            try
            {
                // Check whether authorization header is present
                string auth = Request.Headers.Get("Authorization");
                if (auth == null || auth.Substring(0, 5).ToLower() != "basic")
                    return null;

                string base64Credentials = auth.Substring(6);

                // Decode username and password
                byte[] bytesCredentials = Convert.FromBase64String(base64Credentials);
                string[] credentials = new UTF8Encoding().GetString(bytesCredentials).Split(':');
                string userName = credentials[0];
                string password = credentials[1];

                // Windows Vista sends username in the form of DOMAIN\User
                int delimiterIndex = userName.IndexOf('\\');

                if (delimiterIndex != -1)
                    userName = userName.Remove(0, delimiterIndex + 1);

                IEnumerable<Claim> claims = await Options.OnValidateUser(userName, password, Context);

                // Invalid credentials
                if (claims == null)
                    return null;

                // Authenticated succesefully
                ClaimsIdentity id = new ClaimsIdentity(claims, Options.AuthenticationType);
                return new AuthenticationTicket(id, new AuthenticationProperties());
            }
            catch (Exception ex)
            {
                Logger.Instance.LogError("Failed to authenticate user", ex);
                return null;
            }
        }

        /// <summary>
        /// Handles redirection logic when another type of authentication (i.e. cookie) has failed,
        /// and rewrites "unauthorized" server response with Basic authentication headers.
        /// </summary>
        /// <returns></returns>
        protected override Task ApplyResponseChallengeAsync()
        {
            if (Response.StatusCode == 302)
            {
                string redirectLocation = Response.Headers["Location"];

                if (isLoginUrl(redirectLocation) && isNonBrowserRequest())
                {
                    Logger.Instance.LogDebug(string.Format(
                        "[BasicAuthenticationHandler.ApplyResponseChallenge] RedirectUri: {0}. " +
                        "Issuing Basic authentication", redirectLocation));

                    rewriteUnauthorizedResponse();
                }
                else if (isWebDavRequest())
                {
                    // WebDAV request is sent while ASP.NET Cookie authentication is used. 
                    // XHR request can not display the log-in page or capture 302 response code. 
                    // In this case, instead of "302 Found" we send 278 response code, 
                    // so XHR can process it and redirect to the log-in page.
                    Response.StatusCode = 278; // 278 is an unused status code, it sometimes used instead of 302 for XHR support

                    // Also change the return Url to Referer, as it points to the 
                    // page on which XHR resides, while original Url points to some WebDAV url
                    // Note: we don't use HttpUtility.UrlEncode() method to avoid dependency on System.Web
                    string urlReferrer = Request.Headers["Referer"];
                    Response.Headers["Location"] = string.Format("{0}?{1}={2}",
                        Options.LoginPath, Options.ReturnUrlParameter, Uri.EscapeDataString(urlReferrer));

                    // Allow CORS
                    addCrossDomainHeaders();
                }
            }
            else if (Response.StatusCode == 401)
            {
                addBasicAuthenticationHeader();
            }

            return Task.FromResult<object>(null);
        }

        /// <summary>
        /// Checks whether the supplied URL is a log-in URL used in ASP.NET MVC project.
        /// </summary>
        private bool isLoginUrl(string url)
        {
            if (string.IsNullOrEmpty(url))
                return false;

            int index = url.IndexOf('?');
            if (index >= 0)
                url = url.Substring(0, index);

            return url.EndsWith(Options.LoginPath.ToString(), StringComparison.OrdinalIgnoreCase);
        }

        /// <summary>
        /// Analyzes User-Agent header.
        /// </summary>
        /// <remarks>
        /// If the request is sent by a web browser, the User-Agent header starts 
        /// with 'Mozilla'. Depending on the presence of this string we will use 
        /// Challenge-response authentication or Cookie authentication.
        /// </remarks>
        private bool isNonBrowserRequest()
        {
            string userAgent = Request.Headers["User-Agent"];
            return userAgent == null
                || userAgent.IndexOf("Mozilla", StringComparison.InvariantCultureIgnoreCase) < 0
                || userAgent.IndexOf("Thunderbird", StringComparison.InvariantCultureIgnoreCase) > 0; // Thunderbird User-Agent string strats with "Mozilla".
        }

        /// <summary>
        /// This method is used to distinguish requests sent by XmlHttpRequest 
        /// and by web browser UI, to avoid log-in page redirect for XmlHttpRequest.
        /// </summary>
        private bool isWebDavRequest()
        {
            return Request.Method != "GET" && Request.Method != "HEAD" && Request.Method != "POST";
        }

        /// <summary>
        /// Rewrites "401 Unauthorized" HTTP headers with Basic Authentication header.
        /// </summary>
        private void rewriteUnauthorizedResponse()
        {
            Response.StatusCode = 401;
            Response.ReasonPhrase = "Unauthorized";

            // Null out the Location header from the 302 response so that it
            // doesn't get transmitted to the client unnecessarily.
            Response.Headers["Location"] = String.Empty;

            // Change the response entity to reflect the right message. If we
            // don't clear out the response and re-write it here, then the user
            // might get the wrong message---that is, she might see an HTML page
            // stating that the resource has been moved (which would have been
            // right for 302) when what we want to indicate is an attempt to
            // access an unauthorized resource.
            Response.Body = new MemoryStream();

            addBasicAuthenticationHeader();

            Response.ContentType = "text/html";
            Response.Write(defaultUnauthorizedResponseHtml);
        }

        /// <summary>
        /// Appends "WWW-Authenticate" HTTP header to issue Basic authentication.
        /// </summary>
        private void addBasicAuthenticationHeader()
        {
            Response.Headers.Append("WWW-Authenticate", challenge);
        }

        /// <summary>
        /// Appends HTTP headers to allow cross domain requests.
        /// </summary>
        private void addCrossDomainHeaders()
        {
            string origin = Request.Headers["Origin"];
            Response.Headers.Append("Access-Control-Allow-Origin", String.IsNullOrEmpty(origin) ? "*" : origin.TrimEnd('/'));
            Response.Headers.Append("Access-Control-Allow-Credentials", "true");
            Response.Headers.Append("Access-Control-Expose-Headers", "Location");
            Response.Headers.Append("Access-Control-Max-Age", int.MaxValue.ToString());
        }
    }
}