//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Hosting != HostingEnvironment.IIS)  return string.Empty; #>
//!<# if (!Params.AuthenticationType.HasFlag(AuthenticationType.Digest))  return string.Empty; #>
//!<# if (!Params.IsOwinAuthMode) return string.Empty; #>
using System;
using System.Threading.Tasks;
using Microsoft.AspNet.Identity;

namespace ITHit.WebDav.Server.SampleModel.__Authentication__.DigestAuthentication
{
    /// <summary>
    /// Superclass of ApplicationUserManager that allows to use digest authentication provided by 
    /// <see cref="DigestAuthenticationMiddleware{TKey, TUser, TManager, TLogin, TRole, TClaim}"/>.
    /// </summary>
    /// <typeparam name="TUser">Class that represents user entity (typically <see cref="ApplicationUser"/>).</typeparam>
    /// <typeparam name="TKey">Type of the primary key used by ASP.NET MVC / Entity Framework.</typeparam>
    public class ApplicationUserManagerBase<TUser, TKey> : UserManager<TUser, TKey>
        where TUser : class, IUser<TKey>
        where TKey : IEquatable<TKey>
    {
        /// <summary>
        /// Realm string, used for calculating Digest HA1 hash.
        /// </summary>
        public string Realm
        {
            get
            {
                if (DigestAuthenticationMiddlewareExtensions.DigestAuthenticationOptions == null)
                    return DigestAuthenticationOptions.DefaultRealmString;

                return DigestAuthenticationMiddlewareExtensions.DigestAuthenticationOptions.Realm;
            }
        }

        public ApplicationUserManagerBase(IUserStore<TUser, TKey> store)
            : base(store)
        {
            // Replace password hasher used by ASP.NET Identity with DigestHA1PasswordHasher,
            // which computes Digest HA1 hash string based on username + realm + password,
            // that can be used to verify Digest authentication credentials
            PasswordHasher = new DigestHA1PasswordHasher();
        }

        /// <summary>
        /// Returns string in format of "username:realm:password" which is later used by 
        /// <see cref="DigestHA1PasswordHasher"/> class to computer Digest HA1 hash.
        /// </summary>
        private string getUserRealmAndPassword(string userName, string password)
        {
            return string.Format("{0}:{1}:{2}", userName, Realm, password);
        }

        /// <summary>
        /// Create a user with the given password
        /// </summary>
        public override Task<IdentityResult> CreateAsync(TUser user, string password)
        {
            return base.CreateAsync(user, getUserRealmAndPassword(user.UserName, password));
        }

        /// <summary>
        /// Retrieves the hashed password from the user store and passes username:realm:password string
        /// to <see cref="DigestHA1PasswordHasher.VerifyHashedPassword"/> method.
        /// </summary>
        /// <remarks>
        /// We don't override <see cref="UserManager{TUser}.ChangePasswordAsync"/>, 
        /// <see cref="UserManager{TUser}.CheckPasswordAsync"/> methods since their implementation calls
        /// <see cref="VerifyPasswordAsync"/>, and we don't want username and realm added to the password twice.
        /// </remarks>
        protected override Task<bool> VerifyPasswordAsync(IUserPasswordStore<TUser, TKey> store,
            TUser user, string password)
        {
            return base.VerifyPasswordAsync(store, user, getUserRealmAndPassword(user.UserName, password));
        }

        /// <summary>
        /// Add a user password only if one does not already exist.
        /// </summary>
        public override async Task<IdentityResult> AddPasswordAsync(TKey userId, string password)
        {
            TUser user = await FindByIdAsync(userId);

            if (user == null)
                return IdentityResult.Failed(String.Format("Cannot find user with ID: {0}", userId));

            return await base.AddPasswordAsync(userId, getUserRealmAndPassword(user.UserName, password));
        }

        /// <summary>
        /// Reset a user's password using a reset password token.
        /// </summary>
        public override async Task<IdentityResult> ResetPasswordAsync(TKey userId, string token, string newPassword)
        {
            TUser user = await FindByIdAsync(userId);

            if (user == null)
                return IdentityResult.Failed(String.Format("Cannot find user with ID: {0}", userId));

            return await base.ResetPasswordAsync(userId, token, getUserRealmAndPassword(user.UserName, newPassword));
        }
    }
}