//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Hosting != HostingEnvironment.IIS)  return string.Empty; #>
//!<# if ((!Params.AuthenticationType.HasFlag(AuthenticationType.Digest)))  return string.Empty; #>
//!<# if (!Params.IsOwinAuthMode) return string.Empty; #>
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNet.Identity;
using Microsoft.AspNet.Identity.EntityFramework;
using Microsoft.AspNet.Identity.Owin;
using Microsoft.Owin;
using Microsoft.Owin.Security;
using Microsoft.Owin.Security.Infrastructure;

namespace ITHit.WebDav.Server.SampleModel.__Authentication__.DigestAuthentication
{
    /// <summary>
    /// OWIN middleware that implements 'Digest' authentication protocol.
    /// Implementation itself is in the <see cref="DigestAuthenticationProvider"/> class.
    /// This is just an adapter class so digest authentication is exposed as OWIN Authentication Middleware.
    /// </summary>
    /// <remarks>
    /// Provides fallback to Digest authentication if HTTP response code 302 
    /// (redirect to log-in page) has been issued after (failed) attempt to 
    /// authenticate using other OWIN authentication middleware (i.e. cookie)
    /// that isn't supported by host (i.e. Microsoft Office application).
    /// </remarks>
    internal class DigestAuthenticationHandler<TKey, TUser, TManager, TLogin, TRole, TClaim>
        : AuthenticationHandler<DigestAuthenticationOptions>
        where TKey : IEquatable<TKey>
        where TUser : class, IUser<TKey>
        where TManager : UserManager<TUser, TKey>
        where TLogin : IdentityUserLogin<TKey>
        where TRole : IdentityUserRole<TKey>
        where TClaim : IdentityUserClaim<TKey>
    {
        private const string defaultUnauthorizedResponseHtml =
            @"<html><title>401 Authorization Required</title></html>";

        private DigestAuthenticationProvider providerValue = null;

        /// <summary>
        /// Returns (and initializes as needed) <see cref="DigestAuthenticationProvider"/>.
        /// </summary>
        private DigestAuthenticationProvider Provider
        {
            get
            {
                return providerValue ??
                    (providerValue = new DigestAuthenticationProvider(Context, Options.Realm, getPasswordAndRoles));
            }
        }

        private string challengeValue = null;

        /// <summary>
        /// Returns challenge string returned by server to issue Digest Authentication algorithm.
        /// </summary>
        internal string Challenge
        {
            get
            {
                if (String.IsNullOrWhiteSpace(challengeValue))
                    challengeValue = Provider.GetChallenge();

                return challengeValue;
            }
        }

        public DigestAuthenticationHandler(
            DigestAuthenticationOptions options)
        {
        }

        /// <summary>
        /// Performs request authentication.
        /// </summary>
        /// <returns>
        /// <see cref="AuthenticationTicket"/> class that contains user identity information 
        /// as well as additional authentication state.
        /// </returns>
        protected override Task<AuthenticationTicket> AuthenticateCoreAsync()
        {
            try
            {
                return Task.FromResult(Provider.AuthenticateRequest(Request.Headers, Request.Method));
            }
            catch (Exception ex)
            {
                Logger.Instance.LogError("Failed to authenticate user", ex);
                return null;
            }
        }

        /// <summary>
        /// Handles redirection logic when another type of authentication (i.e. cookie) has failed,
        /// and rewrites "unauthorized" server response with Digest authentication headers.
        /// </summary>
        /// <returns></returns>
        protected override Task ApplyResponseChallengeAsync()
        {
            if (Response.StatusCode == 302)
            {
                string redirectLocation = Response.Headers["Location"];

                if (isLoginUrl(redirectLocation) && isNonBrowserRequest())
                {
                    Logger.Instance.LogDebug(string.Format(
                        "[DigestAuthenticationHandler.ApplyResponseChallenge] RedirectUri: {0}. " +
                        "Issuing Digest authentication", redirectLocation));

                    rewriteUnauthorizedResponse();
                }
                else if (isWebDavRequest())
                {
                    // WebDAV request is sent while ASP.NET Cookie authentication is used. 
                    // XHR request can not display the log-in page or capture 302 response code. 
                    // In this case, instead of "302 Found" we send 278 response code, 
                    // so XHR can process it and redirect to the log-in page.
                    Response.StatusCode = 278; // 278 is an unused status code, it sometimes used instead of 302 for XHR support

                    // Also change the return Url to Referer, as it points to the 
                    // page on which XHR resides, while original Url points to some WebDAV url
                    // Note: we don't use HttpUtility.UrlEncode() method to avoid dependency on System.Web
                    string urlReferrer = Request.Headers["Referer"];
                    Response.Headers["Location"] = string.Format("{0}?{1}={2}",
                        Options.LoginPath, Options.ReturnUrlParameter, Uri.EscapeDataString(urlReferrer));

                    // Allow CORS
                    addCrossDomainHeaders();
                }
            }
            else if (Response.StatusCode == 401)
            {
                addDigestAuthenticationHeader();
            }

            return Task.FromResult<object>(null);
        }

        /// <summary>
        /// Checks whether the supplied URL is a log-in URL used in ASP.NET MVC project.
        /// </summary>
        private bool isLoginUrl(string url)
        {
            if (string.IsNullOrEmpty(url))
                return false;

            int index = url.IndexOf('?');
            if (index >= 0)
                url = url.Substring(0, index);

            return url.ToLowerInvariant().EndsWith(Options.LoginPath.ToString().ToLowerInvariant());
        }

        /// <summary>
        /// Analyzes User-Agent header.
        /// </summary>
        /// <remarks>
        /// If the request is sent by a web browser, the User-Agent header starts 
        /// with 'Mozilla'. Depending on the presence of this string we will use 
        /// Challenge-response authentication or Digest authentication.
        /// </remarks>
        private bool isNonBrowserRequest()
        {
            string userAgent = Request.Headers["User-Agent"];
            return userAgent == null
                || userAgent.IndexOf("Mozilla", StringComparison.InvariantCultureIgnoreCase) < 0
                || userAgent.IndexOf("Thunderbird", StringComparison.InvariantCultureIgnoreCase) > 0; // Thunderbird User-Agent string strats with "Mozilla".
        }

        /// <summary>
        /// This method is used to distinguish requests sent by XmlHttpRequest 
        /// and by web browser UI, to avoid log-in page redirect for XmlHttpRequest.
        /// </summary>
        private bool isWebDavRequest()
        {
            return Request.Method != "GET" && Request.Method != "HEAD" && Request.Method != "POST";
        }

        /// <summary>
        /// Rewrites "401 Unauthorized" HTTP headers with Digest Authentication header.
        /// </summary>
        private void rewriteUnauthorizedResponse()
        {
            Response.StatusCode = 401;
            Response.ReasonPhrase = "Unauthorized";

            // Null out the Location header from the 302 response so that it
            // doesn't get transmitted to the client unnecessarily.
            Response.Headers["Location"] = String.Empty;

            // Change the response entity to reflect the right message. If we
            // don't clear out the response and re-write it here, then the user
            // might get the wrong message---that is, she might see an HTML page
            // stating that the resource has been moved (which would have been
            // right for 302) when what we want to indicate is an attempt to
            // access an unauthorized resource.
            Response.Body = new MemoryStream();

            addDigestAuthenticationHeader();

            Response.ContentType = "text/html";
            Response.Write(defaultUnauthorizedResponseHtml);
        }

        /// <summary>
        /// Appends "WWW-Authenticate" HTTP header to issue Digest authentication.
        /// </summary>
        private void addDigestAuthenticationHeader()
        {
            Response.Headers.Append("WWW-Authenticate", Challenge);
        }

        /// <summary>
        /// Appends HTTP headers to allow cross domain requests.
        /// </summary>
        private void addCrossDomainHeaders()
        {
            string origin = Request.Headers["Origin"];
            Response.Headers.Append("Access-Control-Allow-Origin", String.IsNullOrEmpty(origin) ? "*" : origin.TrimEnd('/'));
            Response.Headers.Append("Access-Control-Allow-Credentials", "true");
            Response.Headers.Append("Access-Control-Expose-Headers", "Location");
            Response.Headers.Append("Access-Control-Max-Age", int.MaxValue.ToString());
        }

        /// <summary>
        /// Returns password and roles for the specified user name
        /// </summary>
        /// <param name="userName">UserName, typically same as e-mail address in default ASP.NET MVC project template.</param>
        /// <param name="context">OWIN context in which this request is being executed.</param>
        /// <returns><see cref="PasswordAndRoles"/>Object containing HA1 password hash and user roles></returns>
        internal static PasswordAndRoles getPasswordAndRoles(string userName, IOwinContext context)
        {
            TManager manager = context.GetUserManager<TManager>();

            IdentityUser<TKey, TLogin, TRole, TClaim> user = manager.FindByName(userName)
                as IdentityUser<TKey, TLogin, TRole, TClaim>;

            if (user != null)
            {
                IList<string> roles = manager.GetRoles(user.Id);
                return new PasswordAndRoles(user.PasswordHash, roles.ToArray());
            }

            return null;
        }
    }
}