//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Hosting != HostingEnvironment.IIS)  return string.Empty; #>
//!<# if (!Params.AuthenticationType.HasFlag(AuthenticationType.Digest))  return string.Empty; #>
//!<# if (!Params.IsOwinAuthMode) return string.Empty; #>
using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Claims;
using System.Security.Cryptography;
using System.Security.Principal;
using System.Text;
using Microsoft.Owin;
using Microsoft.Owin.Security;
using Microsoft.VisualBasic.FileIO;

namespace ITHit.WebDav.Server.SampleModel.__Authentication__.DigestAuthentication
{
    #region class PasswordAndRoles

    /// <summary>
    /// User password and roles.
    /// </summary>
    public class PasswordAndRoles
    {
        /// <summary>
        /// Initializes a new instance of the PasswordAndRoles class.
        /// </summary>
        /// <param name="passwordHash">Password of a user.</param>
        /// <param name="roles">Roles user belongs to.</param>
        public PasswordAndRoles(string passwordHash, string[] roles)
        {
            this.PasswordHash = passwordHash;
            this.Roles = roles;
        }

        /// <summary>
        /// HA1 hash based on username, realm and password.
        /// </summary>
        public string PasswordHash { get; private set; }

        /// <summary>
        /// List of user roles.
        /// </summary>
        public string[] Roles { get; private set; }
    }

    #endregion

    /// <summary>
    /// Contains Digest authentication logic, so it can be reused in ASP.NET digest authentication module
    /// as well as with HTTP listener.
    /// </summary>
    public class DigestAuthenticationProvider
    {
        private bool isNonceStale = false;

        private IOwinContext context;
        private string realm;
        private readonly Func<string, IOwinContext, PasswordAndRoles> getPasswordAndRoles;

        /// <summary>
        /// Initializes a new instance of the DigestAuthenticationProvider class.
        /// </summary>
        /// <param name="context">OWIN context in which this request is being executed.</param>
        /// <param name="realm">Realm string that is typically displayed in the client authenticaiton prompt.</param>
        /// <param name="getPasswordAndRoles">Function which can be used 
        /// to retrieve user's password and roles by his name.</param>
        public DigestAuthenticationProvider(IOwinContext context, string realm,
            Func<string, IOwinContext, PasswordAndRoles> getPasswordAndRoles)
        {
            this.context = context;
            this.realm = realm;
            this.getPasswordAndRoles = getPasswordAndRoles;
        }

        /// <summary>
        /// Whether authorization header is present in headers collection.
        /// </summary>
        /// <param name="headers">Collection with headers.</param>
        /// <returns>'true' if required header is present. 'false' otherwise.</returns>
        public bool IsAuthorizationPresent(IHeaderDictionary headers)
        {
            string authHeader = headers["Authorization"];
            return authHeader != null && authHeader.ToLower().StartsWith("digest");
        }

        /// <summary>
        /// Authenticates request.
        /// </summary>
        /// <param name="headers">Headers collection.</param>
        /// <param name="method">Http verb.</param>
        /// <returns>Authenticated <see cref="IPrincipal"/> instance or <c>null</c> otherwise.</returns>
        public AuthenticationTicket AuthenticateRequest(IHeaderDictionary headers, string method)
        {
            if (!IsAuthorizationPresent(headers))
                return null;

            // 1. Get Username form request
            string auth = headers["Authorization"];
            auth = auth.Trim().Substring(7);

            string[] elements;

            // To avoid any issues when parsing CSV string, such as comas included in quoted fields, we use 
            // TextFieldParser instead of String.Split here.
            using (StringReader reader = new StringReader(auth))
            using (TextFieldParser parser = new TextFieldParser(reader))
            {
                parser.TextFieldType = FieldType.Delimited;
                parser.SetDelimiters(new[] { ",", "=" });
                parser.HasFieldsEnclosedInQuotes = true;
                parser.TrimWhiteSpace = true;
                elements = parser.ReadFields();
            }

            Dictionary<string, string> requestInfo = new Dictionary<string, string>();

            if (elements == null)
                return null;

            for (int i = 0; i < elements.Length; i += 2)
                requestInfo.Add(elements[i], elements[i + 1]);

            string clientUsername = requestInfo.ContainsKey("username")
                ? requestInfo["username"]
                : string.Empty;

            // Workaround for Windows Vista Digest Authorization. 
            // User name may be submitted in the following format: Machine\\User.
            clientUsername = clientUsername.Replace("\\\\", "\\");
            requestInfo["username"] = clientUsername;

            string username = clientUsername;
            int index = username.LastIndexOf('\\');

            if (index > 0)
                username = username.Remove(0, index + 1);

            PasswordAndRoles passwordAndRoles = getPasswordAndRoles(username, context);

            if (passwordAndRoles == null)
                return null;

            // 2. Compute MD5 Digest hash
            string unhashedDigest = GenerateUnhashedDigest(passwordAndRoles.PasswordHash, method,
                requestInfo["uri"], requestInfo["nonce"], requestInfo["qop"], requestInfo["nc"],
                requestInfo["cnonce"]);

            string hashedDigest = CreateMD5HashBinHex(unhashedDigest);

            // 3. Verify credentials by comparing computed hash with hash passed in headers
            // Invalid credentials
            if (requestInfo["response"] != hashedDigest || !isNonceValid(requestInfo["nonce"]))
                return null;

            // Authenticated succesefully
            GenericIdentity id = new GenericIdentity(clientUsername, "digest");

            foreach (string role in passwordAndRoles.Roles)
                id.AddClaim(new Claim(ClaimTypes.Role, role));

            return new AuthenticationTicket(id, new AuthenticationProperties());
        }

        /// <summary>
        /// Returns challenge string that asks the client to issue Digest MD5 authentication.
        /// </summary>
        public string GetChallenge()
        {
            string nonce = createNewNonce();

            StringBuilder stringBuilder = new StringBuilder("Digest");
            stringBuilder.Append(" realm=\"");
            stringBuilder.Append(realm);
            stringBuilder.Append("\", nonce=\"");
            stringBuilder.Append(nonce);
            stringBuilder.Append("\", opaque=\"0000000000000000\", stale=");
            stringBuilder.Append(isNonceStale ? "true" : "false");
            stringBuilder.Append(", algorithm=MD5, qop=\"auth\"");
            return stringBuilder.ToString();
        }

        /// <summary>
        /// Generates unhashed digest string using Digest MD5 algorithm.
        /// </summary>
        /// <param name="ha1">HA1 hash string, calculated based on username, password and realm strings.
        /// You pass PasswordHash field here (which is hashed using our <see cref="DigestHA1PasswordHasher"/> class).</param>
        /// <param name="httpMethod">HTTP method (i.e. "GET" or "POST") used in the request being authenticated.</param>
        /// <param name="uri">Request URI string used in the request being authenticated.</param>
        /// <param name="nonce">Server-generated nonce value used in authentication algorithm.</param>
        /// <param name="qop">"Quality of protection" (qop) parameter used by digest authentication algorithm.</param>
        /// <param name="nc">"Request counter" (nc) parameter used by digest authentication algorithm.</param>
        /// <param name="cnonce">"Client nonce" (cnonce) parameter used by digest authentication algorithm.</param>
        /// <returns>Unhashed Digest string (has to be hashed using CreateMD5HashBinHex method to get Digest MD5 hash).</returns>
        /// <remarks>
        /// See also: http://en.wikipedia.org/wiki/Digest_access_authentication
        /// </remarks>
        internal static string GenerateUnhashedDigest(string ha1, string httpMethod,
            string uri, string nonce, string qop, string nc, string cnonce)
        {
            string a2 = string.Format("{0}:{1}", httpMethod, uri);
            string ha2 = CreateMD5HashBinHex(a2);

            string unhashedDigest;

            if (!String.IsNullOrWhiteSpace(qop))
            {
                unhashedDigest = string.Format("{0}:{1}:{2}:{3}:{4}:{5}",
                    ha1, nonce, nc, cnonce, qop, ha2);
            }
            else
            {
                unhashedDigest = string.Format("{0}:{1}:{2}", ha1, nonce, ha2);
            }

            return unhashedDigest;
        }

        /// <summary>
        /// Returns MD5 hash of the supplied string, formatted as hexadecimal string within curved brackets.
        /// </summary>
        internal static string CreateMD5HashBinHex(string val)
        {
            // If MD5CryptoServiceProvider is created as static member, "handle is invalid" is thrown occasionally
            byte[] ha1Bytes = new MD5CryptoServiceProvider().ComputeHash(Encoding.ASCII.GetBytes(val));
            string ha1 = string.Empty;

            for (int i = 0; i < 16; i++)
                ha1 += string.Format("{0:x02}", ha1Bytes[i]);

            return ha1;
        }

        /// <summary>
        /// Creates new (server-generated) nonce value, used by digest authentication algorithm.
        /// </summary>
        /// <returns></returns>
        private string createNewNonce()
        {
            DateTime nonceTime = DateTime.Now + TimeSpan.FromMinutes(1);
            string expireStr = nonceTime.ToString("G");

            byte[] expireBytes = Encoding.ASCII.GetBytes(expireStr);
            string nonce = Convert.ToBase64String(expireBytes);

            nonce = nonce.TrimEnd(new[] { '=' });
            return nonce;
        }

        /// <summary>
        /// Checks if the passed nonce string is valid 
        /// (by validating the string and verifying nonce expiration date / time).
        /// </summary>
        private bool isNonceValid(string nonce)
        {
            DateTime expireTime;

            int numPadChars = nonce.Length % 4;
            if (numPadChars > 0)
                numPadChars = 4 - numPadChars;

            string newNonce = nonce.PadRight(nonce.Length + numPadChars, '=');

            try
            {
                byte[] decodedBytes = Convert.FromBase64String(newNonce);
                string expireStr = Encoding.ASCII.GetString(decodedBytes);
                expireTime = DateTime.Parse(expireStr);
            }
            catch (FormatException)
            {
                return false;
            }

            return expireTime >= DateTime.Now;
        }
    }
}