//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Hosting != HostingEnvironment.IIS)  return string.Empty; #>
//!<# if (Params.AuthStorage != AuthenticationStorage.Custom)  return string.Empty; #>
//!<# if (!Params.AuthenticationType.HasFlag(AuthenticationType.Msofba))  return string.Empty; #>
//!<# if (!Params.IsOwinAuthMode) return string.Empty; #>
using System.Threading.Tasks;
using Microsoft.Owin;
using Microsoft.Owin.Security;
using Microsoft.Owin.Security.Infrastructure;

namespace ITHit.WebDav.Server.SampleModel.__Authentication__.MSOFBAuthentication
{
    /// <summary>
    /// OWIN middleware that implements 'MS-OFBA' authentication protocol.
    /// </summary>
    /// <remarks>
    /// Provides fallback to MS-OFBA authentication if HTTP response code 302 
    /// (redirect to log-in page) has been issued after (failed) attempt to 
    /// authenticate using other OWIN authentication middleware (i.e. cookie)
    /// that isn't supported by host (i.e. Microsoft Office application).
    /// </remarks>
    public class MSOFBAuthenticationHandler : AuthenticationHandler<MSOFBAuthenticationOptions>
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="MSOFBAuthenticationHandler"/> class.
        /// </summary>
        /// <param name="options">The options.</param>
        public MSOFBAuthenticationHandler(MSOFBAuthenticationOptions options)
        {
        }

        /// <summary>
        /// The core authentication logic which must be provided by the handler. Will be invoked at most
        /// once per request. Do not call directly, call the wrapping Authenticate method instead.
        /// </summary>
        /// <returns>
        /// The ticket data provided by the authentication logic
        /// </returns>
        /// <remarks>
        /// We always return null because goal of this handler is just to tell MS Office to open authentication
        /// window.
        /// </remarks>
        protected override Task<AuthenticationTicket> AuthenticateCoreAsync()
        {
            return Task.FromResult<AuthenticationTicket>(null);
        }

        /// <summary>
        /// Deal with 401 challenge concerns.
        /// </summary>
        /// <reremarks>
        /// If an authentication scheme in question
        /// deals an authentication interaction as part of it's request flow. (like adding a response header, or
        /// changing the 401 result to 302 of a log-in page or external sign-in location.)
        /// </reremarks>
        protected override Task ApplyResponseChallengeAsync()
        {
            if (Request.User == null || !Request.User.Identity.IsAuthenticated)
            {
                string redirectLocation = Response.Headers["Location"];

                Logger.Instance.LogDebug(string.Format(
                    "[MSOFBAuthenticationHandler.ApplyResponseChallenge] RedirectUri: {0}. " +
                    "Issuing MS-OFB authentication", redirectLocation));

                string loginUri = ToAbsolute(Request, Options.LoginPath);
                string successUri = ToAbsolute(Request, Options.LoginSuccessPath);

                Response.StatusCode = 403;
                Response.Headers.Add("X-FORMS_BASED_AUTH_REQUIRED", new[] { string.Format("{0}?{1}={2}", loginUri, Options.ReturnUrlParameter, successUri) });
                Response.Headers.Add("X-FORMS_BASED_AUTH_RETURN_URL", new[] { successUri });
                Response.Headers.Add("X-FORMS_BASED_AUTH_DIALOG_SIZE", new[] { string.Format("{0}x{1}", Options.DialogSize.Width, Options.DialogSize.Height) });
            }

            return Task.FromResult<object>(null);
        }

        /// <summary>
        /// Gets absolute URI from relative.
        /// </summary>
        /// <param name="request">The request.</param>
        /// <param name="relativeUrl">The relative URL.</param>
        private static string ToAbsolute(IOwinRequest request, PathString relativeUrl)
        {
            return string.Format("{0}://{1}{2}", request.Scheme, request.Host.ToUriComponent(), relativeUrl.Value);
        }
    }
}