//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Hosting != HostingEnvironment.IIS)  return string.Empty; #>
//!<# if (!Params.AuthenticationType.HasFlag(AuthenticationType.Msofba))  return string.Empty; #>
//!<# if (!Params.IsOwinAuthMode) return string.Empty; #>
using System;
using System.Threading.Tasks;
using Microsoft.Owin;
using Microsoft.Owin.Security.Infrastructure;

namespace ITHit.WebDav.Server.SampleModel.__Authentication__.MSOFBAuthentication
{
    /// <summary>
    /// Implements MS-OFBA Authentication Middleware.
    /// </summary>
    /// <remarks>
    /// MS-OFBA enables log-in using third party login providers that require to present 
    /// HTML log-in page, such as Facebook, Twitter, Google, etc.
    /// 
    /// MS-OFBA is supported by Microsoft Office 2007 SP1 and later versions.
    /// MS-OFBA is not supported by Microsoft Mini-redirector and OS X Finder.
    /// </remarks>
    internal class MSOFBAuthenticationMiddleware : AuthenticationMiddleware<MSOFBAuthenticationOptions>
    {

        /// <summary>
        /// Initializes a new instance of the <see cref="MSOFBAuthenticationMiddleware"/> class.
        /// </summary>
        /// <param name="next">The next.</param>
        /// <param name="options">The options.</param>
        public MSOFBAuthenticationMiddleware(OwinMiddleware nextMiddleware, MSOFBAuthenticationOptions options)
            : base(nextMiddleware, options)
        {
        }

        /// <summary>
        /// Handles the request.
        /// </summary>
        /// <param name="context">The request context.</param>
        public async override Task Invoke(IOwinContext context)
        {
            // If it's non office request than we should not perform MS-OFBA logic.
            // We can't check it in AuthenticationHandler because base AuthenticationMiddleware
            // add's some logic and office web browser window works incorrectly.
            if (isOFBAAccepted(context.Request))
            {
                await base.Invoke(context);
            }
            else
            {
                await base.Next.Invoke(context);
            }
        }

        /// <summary>
        /// Creates the handler for MS-OFBA Authentication.
        /// </summary>
        /// <returns></returns>
        protected override AuthenticationHandler<MSOFBAuthenticationOptions> CreateHandler()
        {
            return new MSOFBAuthenticationHandler(Options);
        }

        /// <summary>
        /// Analyzes request headers to determine MS-OFBA support.
        /// </summary>
        /// <remarks>
        /// MS-OFBA is supported by Microsoft Office 2007 SP1 and later versions 
        /// and any application that provides X-FORMS_BASED_AUTH_ACCEPTED: t header 
        /// in OPTIONS request.
        /// </remarks>
        private bool isOFBAAccepted(IOwinRequest request)
        {
            // In case application provided X-FORMS_BASED_AUTH_ACCEPTED header
            string ofbaAccepted = request.Headers["X-FORMS_BASED_AUTH_ACCEPTED"];
            if ((ofbaAccepted != null) && ofbaAccepted.Equals("T", StringComparison.InvariantCultureIgnoreCase))
            {
                return true;
            }

            // Microsoft Office does not submit X-FORMS_BASED_AUTH_ACCEPTED header, but it still supports MS-OFBA,
            // Microsoft Office includes "Microsoft Office" string into User-Agent header
            string userAgent = request.Headers["User-Agent"];
            if ((userAgent != null) && userAgent.Contains("Microsoft Office"))
            {
                return true;
            }

            return false;
        }
    }
}