//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Storage != StorageType.Db)  return string.Empty; #>
//!<# if (!Params.DeltaV)  return string.Empty; #>
using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using System.Linq;
using System.Net;
using System.Threading.Tasks;

using ITHit.Server;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Class1;
using ITHit.WebDAV.Server.DeltaV;

namespace ITHit.WebDav.Server.SampleModel.__SqlStorage__
{
    /// <summary>
    /// Represents single file version.
    /// </summary>
    public class Version : IVersion, IContent
    {
        /// <summary>
        /// Instance of <see cref="DavContext"/>.
        /// </summary>
        private readonly DavContext context;
        
        /// <summary>
        /// Number of times corresponding file was modified when version was created.
        /// </summary>
        private readonly string serialNumber;

        /// <summary>
        /// Initializes a new instance of the Version class.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/> class.</param>
        /// <param name="versionId">Id of the version.</param>
        /// <param name="itemId">Id of item version relates to.</param>
        /// <param name="versionNumber">Revision number of the version.</param>
        /// <param name="name">Version name.</param>
        /// <param name="path">Version path.</param>
        /// <param name="created">Time when version was created.</param>
        /// <param name="serialNumber">Number of times corresponding file was modified when version was created.</param>
        public Version(
            DavContext context,
            Guid versionId,
            Guid itemId,
            int versionNumber,
            string name,
            string path,
            DateTime created,
            string serialNumber)
        {
            this.VersionId = versionId;
            this.VersionNumber = versionNumber;

            this.context = context;
            this.ItemId = itemId;
            this.Name = name;
            this.Path = path;
            this.Created = created;
            this.serialNumber = serialNumber;
        }

        //$<IVersion.GetVersionableItemAsync
        /// <summary>
        /// Gets file to which this version relates.
        /// </summary>
        public async Task<IVersionableItem> GetVersionableItemAsync()
        {
                string itemPath = Path.Remove(Path.IndexOf('?'));
                return (IVersionableItem) await context.GetHierarchyItemAsync(itemPath);
        }
        //$>

        //$<IVersion.VersionName
        /// <summary>
        /// Gets display name for the version.
        /// </summary>
        public string VersionName
        {
            get { return "V" + this.VersionNumber; }
        }
        //$>

        /// <summary>
        /// Gets content type of the version.
        /// </summary>
        public string ContentType
        {
            get
            {
                return context.ExecuteScalar<string>(
                    "SELECT ContentType FROM Version WHERE VersionId = @VersionId",
                    "@VersionId", VersionId) ?? string.Empty;
            }
        }

        /// <summary>
        /// Gets content length of the version.
        /// </summary>
        public long ContentLength
        {
            get
            {
                object val = context.ExecuteScalar<object>(
                    "SELECT DATALENGTH(Content) FROM Version WHERE VersionId = @VersionId",
                    "@VersionId", VersionId);

                return val == null ? 0 : Convert.ToInt64(val);
            }
        }

        /// <summary>
        /// Gets version etag.
        /// </summary>
        public string Etag
        {
            get { return string.Format("{0}-{1}", Modified.ToBinary(), serialNumber); }
        }

        /// <summary>
        /// Gets version display name.
        /// </summary>
        public string Name { get; private set; }        

        /// <summary>
        /// Gets date when the version was created.
        /// </summary>
        public DateTime Created { get; private set; }

        /// <summary>
        /// Gets version modification date.
        /// </summary>
        public DateTime Modified
        {
            get { return DateTime.MinValue; }
        }

        /// <summary>
        /// Gets parent folder. Don't need it.
        /// </summary>
        public IFolder Parent
        {
            get { throw new Exception("The method or operation is not implemented."); }
        }

        /// <summary>
        /// Gets version path.
        /// </summary>
        public string Path { get; private set; }

        /// <summary>
        /// Gets revision number.
        /// </summary>
        internal int VersionNumber { get; private set; }        

        /// <summary>
        /// Gets file identifier.
        /// </summary>
        internal Guid ItemId { get; private set; }

        /// <summary>
        /// Gets version identifier.
        /// </summary>
        internal Guid VersionId { get; private set; }

        /// <summary>
        /// Creates path for version for specified file and revision number.
        /// </summary>
        /// <param name="itemPath">Path to the file.</param>
        /// <param name="versionNumber">Revision number.</param>
        /// <returns>Path to the corresponding version.</returns>
        public static string CreateVersionPath(string itemPath, int versionNumber)
        {
            return itemPath + "?version=" + versionNumber;
        }

        //$<IVersion.GetSuccessorAsync
        /// <summary>
        /// Gets next version after this.
        /// </summary>
        /// <returns>Next version or <c>null</c>.</returns>
        public async Task<IVersion> GetSuccessorAsync()
        {
            string command = 
                @"SELECT VersionId, ItemId, VersionNumber, Name, Created, SerialNumber
                  FROM Version
                  WHERE (ItemId = @ItemId) AND (VersionNumber =
                        (SELECT MIN(VersionNumber)
                         FROM Version
                         WHERE (ItemId = @ItemId) AND (VersionNumber > @VersionNumber)))";


            return (await context.ExecuteVersion(
                Path.Remove(Path.IndexOf('?')),
                command,
                "@ItemId", ItemId,
                "@VersionNumber", VersionNumber)).FirstOrDefault();

        }

        //$>

        //$<IVersion.GetPredecessorAsync
        /// <summary>
        /// Gets previous version.
        /// </summary>
        /// <returns>Previous version or <c>null</c>.</returns>
        public async Task<IVersion> GetPredecessorAsync()
        {
            string command = 
                @"SELECT VersionId, ItemId, VersionNumber, Name, Created, SerialNumber
                  FROM Version
                  WHERE (ItemId = @ItemId) AND (VersionNumber =
                        (SELECT MAX(VersionNumber)
                         FROM Version
                         WHERE (ItemId = @ItemId) AND (VersionNumber < @VersionNumber)))";


            return (await context.ExecuteVersion(
                Path.Remove(Path.IndexOf('?')),
                command,
                "@ItemId", ItemId,
                "@VersionNumber", VersionNumber)).FirstOrDefault();
        }

        //$>

        //$<IDeltaVItem.GetCommentAsync
        /// <summary>
        /// Gets comment for this version.
        /// </summary>
        /// <returns>Version comment.</returns>
        public async Task<string> GetCommentAsync()
        {
            return await context.ExecuteScalarAsync<string>(
                "SELECT Comment FROM Version WHERE VersionId = @VersionId",
                "@VersionId", VersionId);
        }
        //$>

        //$<IDeltaVItem.SetCommentAsync
        /// <summary>
        /// Sets comment for this version.
        /// </summary>
        /// <param name="comment">Version comment.</param>
        public async Task SetCommentAsync(string comment)
        {
            await context.ExecuteNonQueryAsync(
                "UPDATE Version SET Comment = @Comment WHERE VersionId = @VersionId",
                "@VersionId", VersionId,
                "@Comment", comment);
        }
        //$>

        //$<IDeltaVItem.GetCreatorDisplayNameAsync
        /// <summary>
        /// Gets name of user who created this version.
        /// </summary>
        /// <returns>Creator name.</returns>
        public async Task<string> GetCreatorDisplayNameAsync()
        {
            return await context.ExecuteScalarAsync<string>(
                "SELECT CreatorDisplayName FROM Version WHERE VersionId = @VersionId",
                "@VersionId", VersionId);
        }
        //$>

        //$<IDeltaVItem.SetCreatorDisplayNameAsync
        /// <summary>
        /// Sets name of the user who created this version.
        /// </summary>
        /// <param name="creatorName">User name.</param>
        public async Task SetCreatorDisplayNameAsync(string creatorName)
        {
            await context.ExecuteNonQueryAsync(
                @"UPDATE Version SET CreatorDisplayName = @CreatorDisplayName
                  WHERE VersionId = @VersionId",
                "@VersionId", VersionId,
                "@CreatorDisplayName", creatorName);
        }
        //$>

        /// <summary>
        /// Reads version content.
        /// </summary>
        /// <param name="output">Stream to write content to.</param>
        /// <param name="startIndex">First byte to read.</param>
        /// <param name="count">Number of bytes to read.</param>
        public async Task ReadAsync(Stream output, long startIndex, long count)
        {
            if (DavFile.ContainsDownloadParam(context.Request.RawUrl))
            {
                DavFile.AddContentDisposition(context, Name);
            }

            await using (SqlDataReader reader = await context.ExecuteReaderAsync(
                CommandBehavior.SequentialAccess,
                "SELECT Content FROM Version WHERE VersionId = @VersionId",
                "@VersionId", VersionId))
            {
                await reader.ReadAsync();

                long bufSize = 1048576; // 1Mb
                byte[] buf = new byte[bufSize];
                long retval;

                while ((retval = reader.GetBytes(
                    reader.GetOrdinal("Content"),
                    startIndex,
                    buf,
                    0,
                    (int)(count > bufSize ? bufSize : count))) > 0)
                {
                    await output.WriteAsync(buf, 0, (int)retval);
                    startIndex += retval;
                    count -= retval;
                }
            }
        }

        /// <summary>
        /// Updating version content is not supported.
        /// </summary>
        public async Task<bool> WriteAsync(Stream content, string contentType, long startIndex, long totalFileSize)
        {
            throw new DavException("Cannot update version.", DavStatus.FORBIDDEN);
        }

        /// <summary>
        /// Retrieves dead properties from the version.
        /// </summary>
        /// <param name="names">Names of properties to be retrieved.</param>
        /// <param name="allprop">Whether all properties shall be retrieved.</param>
        /// <returns>Property values.</returns>
        public async Task<IEnumerable<PropertyValue>> GetPropertiesAsync(IList<PropertyName> names, bool allprop)
        {
            IList<PropertyValue> propVals = await context.ExecutePropertyValueAsync(
                @"SELECT Name, Namespace, PropVal
                  FROM VersionProperty
                  WHERE VersionId = @VersionId",
                "@VersionId", VersionId);

            if (allprop)
            {
                return propVals;
            }
            else
            {
                IList<PropertyValue> requestedPropVals = new List<PropertyValue>();
                foreach (PropertyValue p in propVals)
                {
                    if (names.Contains(p.QualifiedName))
                    {
                        requestedPropVals.Add(p);
                    }
                }
                return requestedPropVals;
            }
        }

        /// <summary>
        /// Gets names of dead properties for this version.
        /// </summary>
        /// <returns>Enumerable with property names.</returns>
        public async Task<IEnumerable<PropertyName>> GetPropertyNamesAsync()
        {
            IList<PropertyName> propNames = new List<PropertyName>();
            foreach (PropertyValue propName in await GetPropertiesAsync(new PropertyName[0], true))
            {
                propNames.Add(propName.QualifiedName);
            }
            return propNames;
        }

        /// <summary>
        /// Copying of versions is not supported.
        /// </summary>
        public async Task CopyToAsync(IItemCollection destFolder, string destName, bool deep, MultistatusException multistatus)
        {
            throw new DavException("Cannot copy version.", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Moving versions is not supported.
        /// </summary>
        public async Task MoveToAsync(IItemCollection destFolder, string destName, MultistatusException multistatus)
        {
            throw new DavException("Cannot move version.", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Delets the version.
        /// </summary>
        /// <param name="multistatus">We don't use it here.</param>
        public async Task DeleteAsync(MultistatusException multistatus)
        {
            await context.ExecuteNonQueryAsync(
                "DELETE FROM Version WHERE VersionId = @VersionId",
                "@VersionId", VersionId);
        }

        /// <summary>
        /// Updating properties is not supported.
        /// </summary>
        public async Task UpdatePropertiesAsync(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {
            throw new DavException("Cannot update version properties", DavStatus.NOT_ALLOWED);
        }
    }
}
