//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Storage != StorageType.Db)  return string.Empty; #>
//!<# if (!Params.DeltaV)  return string.Empty; #>
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.DeltaV;

namespace ITHit.WebDav.Server.SampleModel.__SqlStorage__
{
    /// <summary>
    /// Is used for various operations with a file's versions.
    /// </summary>
    public class VersionHistory : IHistory
    {
        /// <summary>
        /// Instance of <see cref="DavContext"/>.
        /// </summary>
        private readonly DavContext context;

        /// <summary>
        /// Identifier of corresponding file.
        /// </summary>
        private readonly Guid itemId;        

        /// <summary>
        /// Initializes a new instance of the VersionHistory class.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        /// <param name="itemId">File id.</param>
        /// <param name="path">Version history encoded path.</param>
        public VersionHistory(DavContext context, Guid itemId, string path)
        {
            this.context = context;
            this.itemId = itemId;
            this.Path = path;
        }

        /// <summary>
        /// Gets version history path.
        /// </summary>
        public string Path { get; private set; }        

        /// <summary>
        /// Gets display name - not implemented.
        /// </summary>
        public string Name
        {
            get { throw new DavException("The method or operation is not implemented.", DavStatus.NOT_IMPLEMENTED); }
        }

        /// <summary>
        /// Not needed.
        /// </summary>
        public DateTime Created
        {
            get { throw new DavException("The method or operation is not implemented.", DavStatus.NOT_IMPLEMENTED); }
        }

        /// <summary>
        /// Not needed.
        /// </summary>
        public DateTime Modified
        {
            get { throw new DavException("The method or operation is not implemented.", DavStatus.NOT_IMPLEMENTED); }
        }

        //$<IHistory.GetCurrentVersionAsync
        /// <summary>
        /// Gets curret version for the corresponding file.
        /// </summary>
        /// <returns>Instance of <see cref="IVersion"/>.</returns>
        public async Task<IVersion> GetCurrentVersionAsync()
        {
            string command = 
                @"SELECT VersionId, ItemId, VersionNumber, Name, Created, SerialNumber
                  FROM Version
                  WHERE ItemId = @ItemId AND VersionNumber = 
                       (SELECT MAX(VersionNumber) FROM Version WHERE ItemId = @ItemId)";


            return (await context.ExecuteVersion(
                Path.Remove(Path.IndexOf('?')),
                command,
                "@ItemId", itemId)).FirstOrDefault();
        }

        //$>

        //$<IHistory.GetVersionSetAsync
        /// <summary>
        /// Retrieves all versions for corresponding file.
        /// </summary>
        /// <param name="propNames">Properties that will be requested later from the items returned.</param>
        /// <returns>Enumerable with versions.</returns>
        public async Task<IEnumerable<IVersion>> GetVersionSetAsync(IList<PropertyName> propNames)
        {
            string command = 
                @"SELECT VersionId, ItemId, VersionNumber, Name, Created, SerialNumber
                  FROM Version
                  WHERE ItemId = @ItemId";


            return await context.ExecuteVersion(
                Path.Remove(Path.IndexOf('?')),
                command,
                "@ItemId", itemId);
        }

        //$>

        //$<IHistory.GetRootVersionAsync
        /// <summary>
        /// Gets first version.
        /// </summary>
        /// <returns>Instance of <see cref="IVersion"/>.</returns>
        public async Task<IVersion> GetRootVersionAsync()
        {
            string command = 
                @"SELECT VersionId, ItemId, VersionNumber, Name, Created, SerialNumber
                  FROM Version
                  WHERE ItemId = @ItemId AND VersionNumber =
                                   (SELECT MIN(VersionNumber) FROM Version WHERE ItemId = @ItemId)";


            return (await context.ExecuteVersion(
                Path.Remove(Path.IndexOf('?')),
                command,
                "@ItemId", itemId)).FirstOrDefault();
        }
        //$>

        /// <summary>
        /// Not supported.
        /// </summary>
        public async Task CopyToAsync(IItemCollection destFolder, string destName, bool deep, MultistatusException multistatus)
        {
            throw new DavException("Cannot copy history resource", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Not supported.
        /// </summary>
        public async Task MoveToAsync(IItemCollection destFolder, string destName, MultistatusException multistatus)
        {
            throw new DavException("Cannot move history resource", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Not supported.
        /// </summary>
        public async Task DeleteAsync(MultistatusException multistatus)
        {
            throw new DavException("Cannot delete history resource", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Not supported.
        /// </summary>
        /// <param name="names"></param>
        /// <param name="allprop"></param>
        /// <returns></returns>
        public async Task<IEnumerable<PropertyValue>> GetPropertiesAsync(IList<PropertyName> names, bool allprop)
        {
            throw new DavException("Cannot get properties of history resource.", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Not supported.
        /// </summary>
        public async Task<IEnumerable<PropertyName>> GetPropertyNamesAsync()
        {
            throw new DavException("Cannot get property names of history resource.", DavStatus.NOT_ALLOWED);
        }

        /// <summary>
        /// Not supported.
        /// </summary>
        public async Task UpdatePropertiesAsync(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {
            throw new DavException("Cannot update properties of history resource.", DavStatus.NOT_ALLOWED);
        }
    }
}
